//--------------------------------------------------------------------------------------
//
// simple_glow sample
//
// Author: Tristan Lorach
// Email: tlorach@nvidia.com
//
// simple Glow effect, to show how to add a postprocessing effect over
// an existing sample without modifying it too deeply
//
// Copyright (c) NVIDIA Corporation. All rights reserved.
//--------------------------------------------------------------------------------------

#include <stdlib.h>
#include <stdio.h>
#include <ctype.h>
#include <map>

#include <GL/glew.h>
#include <GL/glut.h>

#include <nvGlutManipulators.h>

#define NV_REPORT_COMPILE_ERRORS
#include <nvShaderUtils.h>

#include <Cg/CgGL.h>
#include "FilterBox.h"


static FilterBox*	pFilterBox = NULL;
int					g_mousex = 0, g_mousey = 0;
int					g_width = 0, g_height = 0;
int					g_filterresolution = 256;
int					g_hfiltersize = 128;
float               g_bfactor = 0.8;

using std::map;

////////////////////////////////////////////////////////////////////////////////
//
// Globals
//
////////////////////////////////////////////////////////////////////////////////


nv::GlutExamine manipulator;

enum UIOption {
    OPTION_DISPLAY_WIREFRAME,
    OPTION_LERP_LAYERS,
    OPTION_ANIMATE,
    OPTION_USE_PROGRAM,
    OPTION_COUNT
};
bool options[OPTION_COUNT];
map<char,UIOption> optionKeyMap;

////////////////////////////////////////////////////////////////////////////////
//
// Functions
//
////////////////////////////////////////////////////////////////////////////////


//
//
//////////////////////////////////////////////////////////////////////
void init_opengl() {
    glEnable(GL_DEPTH_TEST);
    glClearColor(0.2, 0.2, 0.2, 1.0);

    glewInit();

    if (!glewIsSupported(
        "GL_VERSION_2_0 "
		"GL_EXT_framebuffer_object "
        "GL_ARB_vertex_program "
        "GL_ARB_fragment_program "
        ))
    {
        printf("Unable to load extensions\n\nExiting...\n");
        exit(-1);
    }
    //
    // Postproc effect
    //
    if(!pFilterBox)
    {
      pFilterBox = new FilterBox();
	  //
	  // here you specify the size of the fbo
	  // where the Gaussian filtering will happen
	  //
      pFilterBox->Initialize(g_filterresolution, g_filterresolution);
    }
}

//
//
//////////////////////////////////////////////////////////////////////
void draw_quad() {
    // r texture coordinate is used to select layer
    glBegin(GL_QUADS);
    glTexCoord3f(0.0, 0.0, 0.0);
    glVertex2f(-1.0, -1.0);
    glTexCoord3f(1.0, 0.0, 4.0);
    glVertex2f(1.0, -1.0);
    glTexCoord3f(1.0, 1.0, 4.0);
    glVertex2f(1.0, 1.0);
    glTexCoord3f(0.0, 1.0, 0.0);
    glVertex2f(-1.0, 1.0);
    glEnd();
}

//
//
//////////////////////////////////////////////////////////////////////
void display() {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    //
    // This part is the basic rendering with no postprocessing
    //
    manipulator.applyTransform();

    glPolygonMode( GL_FRONT_AND_BACK, options[OPTION_DISPLAY_WIREFRAME] ? GL_LINE : GL_FILL);

        glEnable(GL_LIGHTING);
        glEnable(GL_LIGHT0);
        glEnable(GL_COLOR_MATERIAL);
        glEnable(GL_DEPTH_TEST);
		glColor3f(1.0, 0.0, 0.5);
		glutSolidTorus(0.2, 1.0, 20, 30);
		glColor3f(0.3, 0.5, 1.0);
		glutSolidTeapot(0.5f);
		glRotatef(90.0, 1.0, 0,0);
		glColor3f(1.0, 1.0, 0.5);
		glutWireTorus(0.2, 1.0, 20, 30);
        glDisable(GL_LIGHTING);

    glDisable(GL_FRAGMENT_PROGRAM_ARB);

	///////////////////////////////////////////////////////////
    // This part is the additional section to add in order to
    // have the postprocessing effect on top of the basic rendering.
    //
	// Setup of the viewport coords where we want to apply the postprocessing
	// this is made to locally apply the effect instead of doing it in fullscreen only
    //
	pFilterBox->Activate(g_mousex - g_hfiltersize, g_mousey - g_hfiltersize, g_hfiltersize*2, g_hfiltersize*2); 
    //
    // rendering graphics for the postprocessing stage
    // 
	glClearColor(0,0,0,0);
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	glDisable(GL_FRAGMENT_PROGRAM_ARB);
	glDisable(GL_VERTEX_PROGRAM_ARB);
	glPushMatrix();
		glLoadIdentity();
		manipulator.applyTransform();
		glColor3f(1.0, 0.0, 0.5);
		glutWireTorus(0.2, 1.0, 20, 30);
		glColor3f(0.3, 0.5, 1.0);
		glutWireTeapot(0.5f);
	glPopMatrix();
    //
    // Back to the backbuffer
    //
	pFilterBox->Deactivate();
    // Draw the quad on top of the previous rendered scene.
    // using Blend ops
	pFilterBox->Draw(g_bfactor);

	glutSwapBuffers();
}

//
//
//////////////////////////////////////////////////////////////////////
void idle() {
    if ( options[OPTION_ANIMATE])
        manipulator.idle();
    
    glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void key(unsigned char k, int x, int y) {
    k = tolower(k);

    if (optionKeyMap.find(k) != optionKeyMap.end())
        options[optionKeyMap[k]] = !options[optionKeyMap[k]];
	
    switch(k) {
        case 27:
        case 'q':
            exit(0);
            break;
		case '4':
			pFilterBox->Initialize(g_filterresolution, g_filterresolution);
			g_filterresolution += 16;
			printf("filter resolution = %d x %d\n", g_filterresolution, g_filterresolution);
			break;
		case '3':
			pFilterBox->Initialize(g_filterresolution, g_filterresolution);
			g_filterresolution -= 16;
			printf("filter resolution = %d x %d\n", g_filterresolution, g_filterresolution);
			break;
		case '1':
			g_hfiltersize -= 16;
			if(g_hfiltersize < 16) g_hfiltersize = 16;
			printf("filter applied on area %d x %d\n", g_hfiltersize, g_hfiltersize);
			break;
		case '2':
			g_hfiltersize += 16;
			if(g_hfiltersize > 512) g_hfiltersize = 512;
			printf("filter applied on area %d x %d\n", g_hfiltersize, g_hfiltersize);
			break;
        case '=':
        case '+':
            g_bfactor += 0.1f;
            if(g_bfactor > 1.0f)
                g_bfactor = 1.0f;
            break;
        case '-':
        case '_':
            g_bfactor -= 0.1f;
            if(g_bfactor < 0.0f)
                g_bfactor = 0.0f;
            break;
    }
    
	glutPostRedisplay();
}

//
//
//////////////////////////////////////////////////////////////////////
void resize(int w, int h) {
	g_width = w;
	g_height = h;
    glViewport(0, 0, w, h);

	pFilterBox->setWindowViewPort(0, 0, w, h);
    
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    
    gluPerspective(60.0, (GLfloat)w/(GLfloat)h, 0.1, 100.0);
    
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    manipulator.reshape(w, h);
}

//
//
//////////////////////////////////////////////////////////////////////
void mouse(int button, int state, int x, int y) {
    manipulator.mouse(button, state, x, y);
	glutPostRedisplay();
}

void passiveMotion(int x, int y) {
	g_mousex = x;
	g_mousey = g_height - y;
	glutPostRedisplay();
}
//
//
//////////////////////////////////////////////////////////////////////
void motion(int x, int y) {
    manipulator.motion(x, y);
}

void main_menu(int i)
{
  key((unsigned char) i, 0, 0);
}
void init_menus()
{
    int object_menu_id = glutCreateMenu(main_menu);
    {
        glutCreateMenu(main_menu);
        glutAddMenuEntry("' ' - Toggle continuous animation", ' ');
        glutAddMenuEntry("'f' - Toggle using a fragment program", 'f');
        glutAddMenuEntry("'w' - Toggle displaying wireframe", 'w');
        glutAddMenuEntry("'1' - dec area size where we do filtering", '1');
        glutAddMenuEntry("'2' - inc area size where we do filtering", '2');
        glutAddMenuEntry("'3' - dec FBO resolution for filtering", '3');
        glutAddMenuEntry("'4' - inc FBO resolution for filtering", '4');
        glutAddMenuEntry("'-' - dec blend factor", '-');
        glutAddMenuEntry("'+' - inc blend factor", '+');
        glutAddMenuEntry("Quit (esc)", '\033');
        glutAttachMenu(GLUT_RIGHT_BUTTON);
    }
}
//
//
//////////////////////////////////////////////////////////////////////
int main(int argc, char **argv) {
	glutInit(&argc, argv);
	glutInitWindowSize(512, 512);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB);
	glutCreateWindow("simple_glow");

	init_opengl();

    manipulator.setDollyActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_SHIFT);
    manipulator.setPanActivate( GLUT_LEFT_BUTTON, GLUT_ACTIVE_CTRL);
    manipulator.setDollyPosition( -2.0f);

	glutDisplayFunc(display);
	glutPassiveMotionFunc(passiveMotion);
    glutMouseFunc(mouse);
    glutMotionFunc(motion);
    glutIdleFunc(idle);
    glutKeyboardFunc(key);
    glutReshapeFunc(resize);
    init_menus();

    //configure the options
    optionKeyMap['w'] = OPTION_DISPLAY_WIREFRAME;
    options[OPTION_DISPLAY_WIREFRAME] = true;

    optionKeyMap[' '] = OPTION_ANIMATE;
    options[OPTION_ANIMATE] = true;

    //print the help info
    printf( "  Commands:\n");
    printf( "    q / [ESC] - Quit the application\n");
    printf( "    [SPACE]   - Toggle continuous animation\n");
    printf( "    f         - Toggle using a fragment program\n");
    printf( "    w         - Toggle displaying wireframe\n");
    printf( "    1/2       - change area size where we do filtering\n\n");
    printf( "    3/4       - change FBO resolution for filtering\n\n");
    printf( "    +/-       - change blend factor of the glow\n\n");

	glutMainLoop();

	if(pFilterBox) delete pFilterBox; // will call Destroy()
	pFilterBox = NULL;
	return 0;
}
